import io
from contextlib import redirect_stdout
from pathlib import Path

import click

from heinlein import api


@click.command()
@click.argument("dataset_name")
@click.argument("data_type")
@click.argument("path")
@click.option("force", "-f", "--force", is_flag=True)
def add(dataset_name, data_type, path, force=False) -> bool:
    """
    Add a location on disk to a dataset
    """
    api.add(dataset_name, data_type, path, force)
    return True


@click.command()
@click.argument("dataset_name")
@click.argument("data_type")
def remove(dataset_name, data_type):
    api.remove(dataset_name, data_type)
    return True


@click.command()
@click.argument("dataset_name")
@click.argument("data_type")
def get(dataset_name, data_type) -> str:
    """
    Get the path to a specific data type in a specific datset
    """
    # Capture any print statements

    with redirect_stdout(io.StringIO()) as _:
        path = api.get(dataset_name, data_type)

    click.echo(path)


@click.command(name="prep")
@click.argument("dataset_name")
@click.argument("path", type=click.Path(exists=True, path_type=Path))
@click.option(
    "-d",
    "--data-type",
    required=False,
    type=click.Choice(["catalog", "mask", "image"]),
    default="catalog",
)
def prep(dataset_name: str, path: Path, data_type: str = "catalog"):
    """
    Prepare a catalog for a dataset. The path should be a directory containing
    the data as CSV files. The database will be created in the same directory.
    """
    match data_type:
        case "catalog":
            api.prep_catalog(dataset_name, path)
        case _:
            raise NotImplementedError()
    return True



import logging
from pathlib import Path

import numpy as np
from astropy.io import fits

from heinlein.dtypes import mask

from .handler import Handler


def get_mask_handler(path: Path, dconfig: dict):
    try:
        mask_type = dconfig["mask_type"]
    except KeyError:
        raise KeyError(
            "Tried to load this mask but no mask type is specificed in the config file!"
        )

    if mask_type == "fits":
        return FitsMaskHandler(path, dconfig)


class FitsMaskHandler(Handler):
    def __init__(self, path: Path, config: dict, *args, **kwargs):
        super().__init__(path, config, "mask")

    def get_data(self, regions, *args, **kwargs):
        files = [f for f in self._path.glob("*.fits") if not f.name.startswith(".")]
        output = {}
        for name in regions:
            matches = list(filter(lambda x: name in x.name, files))
            if len(matches) > 1:
                logging.error(f"Error: Found more than one mask for region {name}")
                continue

            data = fits.open(matches[0])
            out = np.empty(1, dtype="object")
            out[0] = data
            output.update({name: out})

        return output

    def get_data_object(self, data, *args, **kwargs):
        storage = np.empty(len(data), dtype=object)
        for index, value in enumerate(data.values()):
            storage[index] = value[0]
        return mask.Mask(storage, **self._config)



from collections import OrderedDict
from collections.abc import Iterable
from copy import copy
from functools import singledispatchmethod

from memory_profiler import profile

import heinlein

CURRENT_CACHES = {}


def get_cache(dataset: str):
    max_size = heinlein.get_option("CACHE_SIZE")
    if dataset not in CURRENT_CACHES:
        CURRENT_CACHES[dataset] = Cache(max_size)
    else:
        CURRENT_CACHES[dataset].change_max_size(max_size)
    return CURRENT_CACHES[dataset]


def clear_cache(dataset: str):
    if dataset in CURRENT_CACHES:
        CURRENT_CACHES[dataset].empty()
    else:
        raise ValueError(f"No cache for dataset {dataset} found in memory")


class Cache:
    """
    A caching implementation that attempts to keep track of the amount of memory
    being used. Heinlein data objects must define a method that returns the
    approximate size of an object in bytes. The nice thing is that most of our
    data types are simple arrays of floats at the end of the day.

    The cache thinks in terms of survey regions. A single survey region may
    have multiple objects associated with it. When the cache evicts objects,
    it evicts all objects associated with a given survey region with a LRU
    policy.

    The cache assumes immutability of the underlying data. This means that if
    it recieves a request to add a piece of data that is already in the cache,
    it will raise an error.
    """

    def __init__(self, max_size: float = 4e9):
        self.max_size = max_size  # default to 4 GB
        self.cache = OrderedDict()
        self.sizes = {}
        self.size = 0
        self.ref_counts = {}

    def add(self, data):
        """
        Add objects to the cache. If the cache is full, evict objects until the
        object can be added.
        """
        data_to_cache = switch_major_key(data)
        total_size = 0
        to_update = {}
        new_refs = copy(self.ref_counts)
        objects_to_store = set()
        for region_name, region_data in data_to_cache.items():
            to_update[region_name] = {}
            for dtype, data in region_data.items():
                if self.has_data(region_name, dtype):
                    raise ValueError(
                        "Cannot add data to the cache that is already in the cache"
                    )
                if id(data) not in objects_to_store:
                    objects_to_store.add(id(data))

                nrefs = new_refs.get(id(data), 0)
                if nrefs == 0:
                    total_size += data.estimate_size()
                new_refs[id(data)] = nrefs + 1
                to_update[region_name][dtype] = data
        self.make_space(total_size)

        for region_name, region_data in to_update.items():
            try:
                self.cache[region_name].update(region_data)
            except KeyError:
                self.cache[region_name] = region_data
            self.sizes[region_name] = sum(
                [data.estimate_size() for data in region_data.values()]
            )
        self.size += total_size
        self.ref_counts = new_refs

    def change_max_size(self, new_size: float):
        if new_size > self.size:
            self.max_size = new_size
            return True
        else:
            raise ValueError(
                "Cannot change the cache size to a value smaller than the current size"
                f". Requested: {new_size}, Current size: {self.size}/{self.max_size}"
            )

    def make_space(self, needed_space: int):
        if needed_space > self.max_size:
            raise MemoryError(
                "Cannot make space for an object larger than the cache itself"
                f". Requested: {needed_space}, Cache size: {self.max_size}"
            )

        current_space = self.max_size - self.size
        if needed_space < current_space:
            return True  # we have enough space

        space_to_free = needed_space - current_space
        while space_to_free > 0:
            region_name, region_data = self.cache.popitem()
            region_space = self.sizes.pop(region_name)
            data_ids = [id(data) for data in region_data.values()]
            for did in data_ids:
                self.ref_counts[did] -= 1
                if self.ref_counts[did] == 0:
                    space_to_free -= region_space
                    self.size -= region_space
                    del region_data
        return True

    @profile
    def empty(self):
        self.cache = OrderedDict()
        self.sizes = {}
        self.size = 0

    def drop(self, region_names):
        for region_name in region_names:
            if region_name in self.cache:
                region_space = self.sizes.pop(region_name)
                self.size -= region_space
                del self.cache[region_name]

    @singledispatchmethod
    def get(self, region_name: str, dtypes: Iterable):
        if region_name not in self.cache:
            raise KeyError("Region not in cache")
        if isinstance(dtypes, str):
            dtypes = {dtypes}
        elif not isinstance(dtypes, set):
            dtypes = set(dtypes)

        dtypes_in_cache = set(self.cache[region_name].keys())
        dtypes_to_get = dtypes.intersection(dtypes_in_cache)
        if not dtypes_to_get:
            raise ValueError("None of the requested objects are in the cache")

        self.cache.move_to_end(region_name, last=False)
        return {dtype: self.cache[region_name][dtype] for dtype in dtypes_to_get}

    @get.register
    def _(self, regions: set, dtypes: Iterable):
        output = {}
        for region in regions:
            try:
                output[region] = self.get(region, dtypes)
            except KeyError:
                continue
        return switch_major_key(output)

    def has_data(self, region_name, dtype):
        return region_name in self.cache and dtype in self.cache[region_name]


def switch_major_key(data: dict):
    """
    The godata cache thinks in terms of regions, while the rest of the codebase
    thinks in terms of data types. This function switch the major key from one to
    the other.
    """

    output = {}
    for major_key, minor_data in data.items():
        for minor_key, data in minor_data.items():
            if minor_key not in output:
                output[minor_key] = {}
            output[minor_key][major_key] = data
    return output



from abc import ABC, abstractmethod
from pathlib import Path

from heinlein import dtypes


class Handler(ABC):
    def __init__(self, path: Path, dconfig: dict, type: str, *args, **kwargs):
        self._path = path
        self._type = type
        self._config = dconfig

    @abstractmethod
    def get_data(self, regions: list, *args, **kwargs):
        pass

    def get_data_object(self, data, *args, **kwargs):
        return dtypes.get_data_object(self._type, data)



class HeinleinError(Exception):
    pass


class HeinleinConfigError(HeinleinError):
    @classmethod
    def doesnt_exist(cls, option: str) -> "HeinleinConfigError":
        return cls(f"'{option}' is not a valid configuration option")



from __future__ import annotations

from abc import ABC, abstractclassmethod, abstractmethod

from heinlein.region import BaseRegion


class HeinleinDataObject(ABC):
    """
    This is a generic class that is used for storing data objects in the cache. An
    object of this type will never actually be given to the user. The data objects
    must define a `get_data_from_region` method, which will return the data from the
    region provided.

    """

    @abstractmethod
    def get_data_from_region(self, region: BaseRegion):
        pass

    @abstractclassmethod
    def combine(cls, objects: list[HeinleinDataObject]):
        pass

    @abstractmethod
    def estimate_size(self) -> int:
        """
        estimate the size of the object in bytes
        """



from math import ceil

import numpy as np
from astropy.coordinates import SkyCoord


def warning_prompt(warning: str, options: list) -> str:
    print(warning)
    keys = [opt[0].upper() for opt in options]
    while True:
        for index, option in enumerate(options):
            print(f"{option} ({keys[index]})")
        i = input("?: ")
        if i.upper() in keys:
            return i.upper()
        else:
            print("Invalid option")


def warning_prompt_tf(warning: str) -> bool:
    options = ["Yes", "No"]
    if warning_prompt(warning, options) == "Y":
        return True
    return False


def initialize_grid(bounds, area, density):
    ra1, ra2 = bounds[0], bounds[2]
    dec1, dec2 = bounds[1], bounds[3]
    dra = ra2 - ra1
    ddec = dec2 - dec1
    npoints = round(density * area)
    ratio = ddec / dra
    nra = np.sqrt(npoints / ratio)
    ndec = nra * ratio
    nra = ceil(nra) + 1
    ndec = ceil(ndec) + 1
    dra_ = dra / (nra + 1)
    ddec_ = ddec / (ndec + 1)
    gra = [ra1 + i * dra_ for i in range(1, nra + 1)]
    gdec = [dec1 + i * ddec_ for i in range(1, ndec + 1)]
    x_, y_ = np.meshgrid(gra, gdec)
    cpoints = np.vstack([x_.ravel(), y_.ravel()])
    coords = SkyCoord(cpoints[0], cpoints[1], unit="deg")
    return coords



from pathlib import Path

import pandas as pd
from astropy.io import ascii
from astropy.table import Table
from sqlalchemy import create_engine, text

from heinlein.dtypes.catalog import Catalog
from heinlein.dtypes.handlers import handler


class heinleinIoException(Exception):
    pass


def get_catalog_handler(config: dict, dconfig: dict):
    data = config.get("data", {})
    try:
        path = Path(data["catalog"])
    except KeyError:
        raise heinleinIoException("No catalog path found in config!")

    if path.is_dir():
        return CsvCatalogHandler(path, dconfig)

    elif path.suffix == ".db":
        return SQLiteCatalogHandler(path, dconfig)
    else:
        raise heinleinIoException("Catalog path is not a directory or a .db file!")


class CsvCatalogHandler(handler.Handler):
    def __init__(self, path: Path, config: dict, *args, **kwargs):
        super().__init__(path, config, "catalog")
        self.known_files = [f for f in self._path.glob("*") if f.is_file()]

    def get_data(self, region_names: list, *args, **kwargs):
        """
        Default handler for a catalog.
        Loads a single catalog, assuming the region name can be found in the file name.
        """
        storage = {}
        files = {}
        for name in region_names:
            region_file = list(filter(lambda x: name in x.name, self.known_files))
            if len(region_file) == 0:
                raise heinleinIoException(
                    "No file found for region "
                    f"{name}! Perhaps it is in an external"
                    " storage device that isn't attached."
                )
            elif len(region_file) != 1:
                raise heinleinIoException(
                    "Multiple files found for region {name}! Found {region_file}"
                )
            files.update({name: region_file[0]})

        for name, path in files.items():
            if not path.exists():
                print(
                    f"Path {self._path} does not exist! Perhaps it is "
                    "in an external storage device that isn't attached."
                )
                return None
            data = ascii.read(path)
            storage.update({name: Catalog(data, self._config)})
        return storage


class SQLiteCatalogHandler(handler.Handler):
    def __init__(self, path: Path, config: dict, *args, **kwargs):
        super().__init__(path, config, "catalog")
        if not path.exists():
            raise heinleinIoException(f"Path {path} does not exist!")
        elif path.suffix != ".db":
            raise heinleinIoException(f"Path {path} is not a .db file!")

        self._create_engine()

    def _create_engine(self, *args, **kwargs):
        self._engine = create_engine(f"sqlite:///{self._path}")
        self._con = self._engine.connect()
        sql = "SELECT * FROM sqlite_master where type='table'"
        cur = self._con.execute(text(sql))
        self._tnames = [t[1] for t in cur.fetchall()]

    def get_data(self, region_names: list, *args, **kwargs):
        subregion_key = self._config.get("subregion", None)
        if subregion_key is not None:
            splits = [rname.split(".") for rname in region_names]
            regions_to_get = {}
            for split in splits:
                if len(split) == 2:
                    if split[0] in regions_to_get.keys():
                        regions_to_get[split[0]].append(split[1])
                    else:
                        regions_to_get.update({split[0]: [split[1]]})
                elif len(split) == 2 and split[0] not in regions_to_get.keys():
                    regions_to_get.update({split[0]: []})

            storage = self.get_with_subregions(regions_to_get)
        else:
            storage = self._get(region_names)

        return {k: Catalog(table, self._config) for k, table in storage.items()}

    def get_with_subregions(self, regions: dict):
        subregion_key = self._config.get("subregion", None)
        region_key = self._config.get("region", None)
        storage = {}
        for region, subregions in regions.items():
            region_names = [".".join([region, sr]) for sr in subregions]
            if str(region) in self._tnames:
                table = self.get_where(region, {subregion_key: subregions})
                if len(table) != 0:
                    for index, sr in enumerate(subregions):
                        mask = (table[region_key].astype(str) == region) & (
                            table[subregion_key].astype(str) == sr
                        )
                        storage.update({region_names[index]: table[mask]})
                else:
                    storage.update({sr: Catalog() for sr in region_names})
                    # This needs to be fixed
            elif len(self._tnames) == 1:
                region_names = [".".join([region, sr]) for sr in subregions]

                table = self.get_where(
                    self._tnames[0],
                    {region_key: region.name, subregion_key: subregions},
                )
                for index, sr in enumerate(subregions):
                    mask = table[subregion_key] == sr
                    storage.update({region_names[index]: table[mask]})
            else:
                storage.update({sr: Table() for sr in region_names})
        return storage

    def _get(self, region_names: list):
        storage = {}
        for region in region_names:
            if region in self._tnames:
                table = self.get_all(region)

                storage.update({region: table})

            elif len(self._tnames) == 1:
                region_key = self._config.get("region", None)
                table = self.get_where(self._tnames[0], {region_key: region.name})
                storage.update({region: table})
        return storage

    def get_where(self, tname: str, conditions: dict):
        base_query = f'SELECT * FROM "{tname}" WHERE '
        base_condition = '{} = "{}"'
        multiple_base_conditions = "{} IN ({})"
        output_conditions = []
        for k, vs in conditions.items():
            if isinstance(vs, list):
                c = '","'.join([str(v) for v in vs])
                c = '"' + c + '"'
                output_conditions.append(multiple_base_conditions.format(k, c))
            else:
                output_conditions.append(base_condition.format(k, vs))
        query = base_query + " AND ".join(output_conditions)
        table = self.execute_query(query)
        return self._parse_return(table)

    def get_all(self, tname):
        query = f'SELECT * FROM "{tname}"'
        table = self.execute_query(query)
        return self._parse_return(table)

    def execute_query(self, query):
        q = text(query)
        data = pd.read_sql(q, self._con)
        return data

    def _parse_return(self, data, *args, **kwargs):
        if len(data) == 0:
            return Catalog()

        # replace all None with np.nan
        data = Table.from_pandas(data)
        return data



from __future__ import annotations

import json
import logging
from abc import ABC, abstractmethod
from typing import Any

from astropy.coordinates import SkyCoord
from spherical_geometry.polygon import SingleSphericalPolygon

from heinlein.locations import MAIN_CONFIG_DIR

logger = logging.getLogger("region")


def load_config(*args, **kwargs):
    """
    Loads the region config.
    """
    config_location = MAIN_CONFIG_DIR / "region" / "region.json"
    with open(config_location, "r") as f:
        config = json.load(f)
        return config


def create_bounding_box(bounds: tuple) -> SingleSphericalPolygon:
    """
    Create a SingleSphericalPolygon from bounds, setup as "RA1 DEC1 RA2 DEC2"
    Note, this does NOT check that the ra_min < ra_max or dec_min < dec_max,
    because some regions may straddle the 0/360 line or the poles.
    """

    ra_min, dec_min, ra_max, dec_max = bounds
    if ra_min == ra_max or dec_min == dec_max:
        raise ValueError("Invalid bounds: Box has zero width or height")
    RA_STRADDLES_ZERO = ra_min > ra_max
    DEC_STRADDLES_POLE = dec_min > dec_max
    ra_mid = (ra_min + ra_max) / 2
    dec_mid = (dec_min + dec_max) / 2
    if RA_STRADDLES_ZERO:
        ra_mid = (ra_mid + 180) % 360
    if DEC_STRADDLES_POLE:
        dec_mid = (dec_mid + 90) % 180
    ras = [ra_min, ra_max, ra_max, ra_min]
    decs = [dec_min, dec_min, dec_max, dec_max]
    return SingleSphericalPolygon.from_radec(
        ras, decs, center=(ra_mid, dec_mid), degrees=True
    )


current_config = load_config()


class BaseRegion(ABC):
    _config = current_config

    def __init__(
        self,
        polygon: SingleSphericalPolygon,
        bounds: tuple,
        name=None,
        regtype=None,
        *args,
        **kwargs,
    ):
        """
        Base region object.
        Regions should always be initialized with heinlein.Region

        parameters:

        geometry: <spherical_geometry.SphericalPolygon> The sky region
        type: <str> The type of the region
        name: <str> The name of the region (optional)
        """
        self.spherical_geometry = polygon
        self.bounding_box = create_bounding_box(bounds)
        self.name = name
        self._type = regtype

    def __getattribute__(self, __name: str) -> Any:
        """
        Implements geometry relations for regions. Delegates unknown methods to the
        underlying spherical geometry object, IF that method is explicitly permitted
        in heinlein/config/region/region.json
        """
        try:
            config = object.__getattribute__(self, "_config")
            cmd_name = config["allowed_predicates"][__name]
            sg = object.__getattribute__(self, "spherical_geometry")
            predicate = getattr(sg, cmd_name)

            def do_predicate(other: BaseRegion):
                return predicate(other.spherical_geometry)

            return do_predicate
        except (AttributeError, KeyError):
            return object.__getattribute__(self, __name)

    @property
    def bounds(self):
        ra, dec = self.bounding_box.to_lonlat()
        min_ra, max_ra = ra[0], ra[2]
        min_dec, max_dec = dec[0], dec[1]
        return min_ra, min_dec, max_ra, max_dec

    @abstractmethod
    def contains(self, point: SkyCoord) -> bool:
        """
        Check if a point is in the region
        """
        pass



from pathlib import Path

from heinlein.manager.manager import get_manager, initialize_dataset
from heinlein.utilities import prep, warning_prompt_tf

"""
Backend API functions
"""


def add(name, dtype, path, force=False, *args, **kwargs) -> bool:
    """
    Add a location on disk to a dataset
    """
    cwd = Path.cwd()

    if path != "cwd":
        path = cwd / path
    else:
        path = cwd
    if not path.exists():
        print(f"Error: {path} not found!")
        return

    try:
        data_manager = get_manager(name)
    except FileNotFoundError:
        if not force:
            write_new = warning_prompt_tf(
                f"Survey {name} not found, would you like to initialize it? "
            )
            if not write_new:
                print("Aborting...")
                return
        initialize_dataset(name, path)
        data_manager = get_manager(name)

    data_manager.add_data(dtype, path)
    return True


def remove(name: str, dtype: str):
    """
    Remove a datatype from a dataset
    """
    try:
        manager = get_manager(name)
    except FileNotFoundError:
        print(f"Error: dataset {name} does not exist!")
        return True

    manager.remove_data(dtype)
    return True


def get(name: str, dtype: str) -> Path:
    """
    Get the path to a specific data type in a specific datset
    """
    try:
        manager = get_manager(name)
    except FileNotFoundError:
        print(f"Error: dataset {name} does not exist!")
    path = manager.get_path(dtype)

    return path


def prep_catalog(name: str, path: Path):
    """
    Prepare a catalog for a dataset. The path should be a directory containing
    the data as CSV files. The database will be created in the same directory.
    """
    if path.suffix == ".sqlite3":
        prep.register_database(name, path)

    csvs = list(path.glob("*.csv"))
    if not csvs:
        raise FileNotFoundError(f"No CSV files found in {path}")
    prep.database_from_csvs(name, csvs)



from __future__ import annotations

import json
from functools import cache, reduce
from operator import add

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import Table, vstack
from spherical_geometry.vector import lonlat_to_vector

from heinlein.dtypes import dobj
from heinlein.locations import MAIN_CONFIG_DIR
from heinlein.region import BaseRegion


def load_config():
    catalog_config_location = MAIN_CONFIG_DIR / "dtypes" / "catalog.json"
    with open(catalog_config_location, "rb") as f:
        data = json.load(f)
    return data


def get_coordinates(catalog: Table):
    if "coordinates" in catalog.colnames and isinstance(
        catalog["coordinates"], SkyCoord
    ):
        return catalog["coordinates"]
    coords = SkyCoord(catalog["ra"], catalog["dec"])
    return coords


def get_cartesian_points(coordinates: SkyCoord):
    lon = coordinates.ra.to_value("deg")
    lat = coordinates.dec.to_value("deg")
    cartesian_points = np.dstack(lonlat_to_vector(lon, lat))[0]
    return cartesian_points


def label_coordinates(catalog: Table, config: dict = {}):
    """
    Takes a catalog and finds the coordinate columns (ra and dec)
    returns the catalog with the coordinate columns labeled
    """

    config_has_coord_columns = config.get("columns", False) and config["columns"].get(
        "ra", False
    )
    if not config_has_coord_columns:
        default_config = load_config()
        columns = set(catalog.colnames)
        ras = set(default_config["columns"]["ra"])
        dec = set(default_config["columns"]["dec"])
        ra_col = columns.intersection(ras)
        dec_col = columns.intersection(dec)
        if len(ra_col) != 1 or len(dec_col) != 1:
            raise ValueError("Catalog does not have the correct columns for ra and dec")
        ra_name = list(ra_col)[0]
        dec_name = list(dec_col)[0]
        ra_unit = u.deg
        dec_unit = u.deg
    else:
        ra_name = config["columns"]["ra"]["key"]
        dec_name = config["columns"]["dec"]["key"]
        ra_unit_name = config["columns"]["ra"].get("unit", "deg")
        dec_unit_name = config["columns"]["dec"].get("unit", "deg")
        ra_unit = getattr(u, ra_unit_name)
        dec_unit = getattr(u, dec_unit_name)

    catalog.rename_columns([ra_name, dec_name], ["ra", "dec"])
    catalog._has_radec = True
    try:
        catalog["ra"].to(u.deg)
    except u.UnitConversionError:
        catalog["ra"] = catalog["ra"] * ra_unit
    try:
        catalog["dec"].to(u.deg)
    except u.UnitConversionError:
        catalog["dec"] = catalog["dec"] * dec_unit

    return catalog


def Catalog(data: Table = None, config: dict = {}, *args, **kwargs):
    """
    Produces a catalog object from an astropy table. This locates the ra and dec
    columns, creates SkyCoords, and creates a cartesian representation of the
    coordinates for more efficient filtering. Produces a :class:`CatalogObject` which
    is used internally by the DataManager. When a pieces of data is actuall requested
    this object will produce a :class:`astropy.Table` with the data from the region.

    """
    if data is None:
        data = Table()
    if len(data) == 0:
        points = np.array([])
        return CatalogObject(data, points)

    labeled_data = label_coordinates(data, config)
    catalog_coordinates = get_coordinates(labeled_data)
    data["coordinates"] = catalog_coordinates
    return CatalogObject(data)


class CatalogObject(dobj.HeinleinDataObject):
    def __init__(self, data: Table):
        self._data = data

    def __len__(self):
        return len(self._data)

    def get_data_from_region(self, region: BaseRegion):
        mask = region.contains(self._data["coordinates"])
        return self._data[mask]

    @classmethod
    def combine(cls, objects: list[CatalogObject]):
        data = vstack([o._data for o in objects if len(o._data) > 0])
        return cls(data)

    @cache
    def estimate_size(self) -> int:
        data_size = reduce(
            add, [estimate_column_size(self._data[col]) for col in self._data.colnames]
        )
        return data_size


def estimate_column_size(column: np.ndarray) -> int:
    if isinstance(column, SkyCoord):
        return column.size * 16
    return column.nbytes



from __future__ import annotations

import json
import logging
import multiprocessing as mp
from functools import cache
from importlib import import_module
from inspect import getmembers, isclass, isfunction
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Optional

import appdirs

from heinlein.errors import HeinleinError
from heinlein.manager.cache import get_cache
from heinlein.region.base import BaseRegion
from heinlein.utilities import warning_prompt


class MissingDataError(HeinleinError):
    pass


logger = logging.getLogger("manager")
known_datasets = ["des", "cfht", "hsc", "ms"]
active_managers = {}


def get_manager(name: str) -> DataManager:
    try:
        am = active_managers[name]
        return am
    except KeyError:
        mgr = DataManager(name)
        active_managers.update({name: mgr})
        return mgr


@cache
def get_config_location() -> Path:
    return Path(appdirs.user_config_dir("heinlein"))


def get_dataset_config_dir(dataset_name: str) -> Path:
    return get_config_location() / dataset_name


def initialize_dataset(name: str, *args, **kwargs):
    """
    Initializes a new dataset. Checks to see if a custom implementation exists for a
    given dataset. These have to be installed separately from the main package.
    """
    print(f"Initializing dataset {name}...")
    ext = get_external_implementation(name)
    config_data = ext.load_config()
    config_location = get_dataset_config_dir(name)
    config_location.mkdir(parents=True, exist_ok=True)
    config_data.update({"data": {}})
    with open(config_location / "config.json", "w") as f:
        json.dump(config_data, f)


def get_external_implementation(name: str) -> Optional[ModuleType]:
    """
    Checks to see if a custom implementation exists for a given dataset. These have
    to be installed separately from the main package. Prompts the user to install
    the package if it is not found, but is known.
    """
    if name not in known_datasets:
        raise ValueError(f"Dataset {name} is not supported!")
    else:
        # try to import it
        module_name = "heinlein_" + name
        try:
            module = import_module(module_name)
        except ImportError:
            print(
                f"Dataset `{name}` is a known dataset, but needs to be installed"
                f" separately. You can install it with `pip install heinlein_{name}`."
            )
            return None
        return getattr(module, name)


def check_overload(f):
    def wrapper(self, *args, **kwargs):
        bypass = kwargs.get("bypass", False)
        if self.external is None or bypass:
            return f(self, *args, **kwargs)
        else:
            try:
                ext_fn = self._external_definitions[f.__name__]
                return ext_fn(self, *args, **kwargs)
            except KeyError:
                return f(self, *args, **kwargs)

    return wrapper


def get_dataset_config(dataset_name: str) -> dict:
    config_location = get_dataset_config_dir(dataset_name)
    if not config_location.exists():
        raise FileNotFoundError(f"Dataset {dataset_name} has not been initialized!")
    with open(config_location / "config.json", "r") as f:
        config_data = json.load(f)
    return config_data


def write_dataset_config(dataset_name: str, config_data: dict) -> None:
    config_location = get_dataset_config_dir(dataset_name)
    if not config_location.exists():
        raise FileNotFoundError(f"Dataset {dataset_name} has not been initialized!")
    with open(config_location / "config.json", "w") as f:
        json.dump(config_data, f)


logger = logging.getLogger("manager")


class DataManager:
    def __init__(self, name: str, *args, **kwargs):
        """
        The datamanger just tracks where data and configuration are for
        a given dataset.

        parameters:

        name: <str> The name of the dataset
        """
        self.name = name
        self._setup()
        self._cache_lock = mp.Lock()

    def _setup(self, *args, **kwargs) -> None:
        """
        Performs basic setup of the manager
        Loads datset config if it exists, or prompts
        user if dataset does not exist.
        """
        try:
            # Find the external implementation for this dataset, if it exists.
            self.external = get_external_implementation(self.name)
        except KeyError:
            self.external = None
        self._initialize_external_implementation()

        self.config = get_dataset_config(self.name)

    def _initialize_external_implementation(self):
        """
        Should be something like "load plugins"
        """
        if self.external is None:
            self._external_definitions = {}
            return
        fns = [f for f in getmembers(self.external, isfunction)]
        fns = list(
            filter(lambda f, m=self.external: f[1].__module__ == m.__name__, fns)
        )
        external_functions = {fn[0]: fn[1] for fn in fns}

        classes = [f for f in getmembers(self.external, isclass)]
        classes = list(
            filter(lambda f, m=self.external: f[1].__module__ == m.__name__, classes)
        )
        external_classes = {cl[0]: cl[1] for cl in classes}

        fn_keys = set(external_functions.keys())
        cls_keys = set(external_classes.keys())
        if len(fn_keys.intersection(cls_keys)) != 0:
            print(
                "Error: Overloaded functions and classes in dataset implementations"
                "Should all have unique names, but found duplicates for {keys_}"
            )
            exit()
        self._external_definitions = {**external_classes, **external_functions}

    def get_external(self, key: str, *args, **kwargs) -> Optional[Callable]:
        return self._external_definitions.get(key, None)

    def get_path(self, dtype: str, *args, **kwargs) -> Path:
        data = self.config.get("data", {})
        if dtype not in data:
            raise KeyError(f"Datatype {dtype} not found for dataset {self.name}!")
        return Path(data[dtype])

    def load_handlers(self, *args, **kwargs):
        from heinlein.dtypes import handlers

        if not hasattr(self, "_handlers"):
            data = self.config.get("data", {})
            known_dtypes = list(data.keys())
            self._handlers = handlers.get_file_handlers(
                known_dtypes, self.config, self._external_definitions
            )

    @staticmethod
    def exists(name: str) -> bool:
        """
        Checks if a datset exists
        """
        try:
            get_dataset_config(name)
            return True
        except FileNotFoundError:
            return False

    def add_data(self, dtype: str, path: Path, overwrite=False) -> bool:
        """
        Add data to a datset. Note that this only gives the manager
        a path to the data. The manager itself does not know what kind
        of data it is or how to use it. Usually this will be invoked
        by a command line script.

        Params:

        dtype <str>: Type of data being added (i.e. "catalog")
        path <pathlib.Path>: Path to the data

        Returns:

        bool: Whether or not the file was sucessfully added
        """
        data = self.config.get("data", {})
        if dtype in data and not overwrite:
            msg = f"Datatype {dtype} already found for survey {self.name}."
            options = ["Overwrite", "Merge", "Abort"]
            choice = warning_prompt(msg, options)
            if choice == "A":
                return False
            elif choice == "M":
                raise NotImplementedError
        if not path.exists():
            print(f"Error: File {path} does not exist!")
            return False

        data.update({dtype: str(path)})
        self.config.update({"data": data})
        write_dataset_config(self.name, self.config)
        return True

    def remove_data(self, dtype: str) -> bool:
        """
        Remove data from a datset. Usually this will be invoked
        by a command line script.

        Params:

        dtype <str>: Type of data being added (i.e. "catalog")

        Returns:

        bool: Whether or not the file was sucessfully removed
        """
        data = self.config.get("data", {})
        if dtype not in data:
            print(f"Error: Datatype {dtype} not found for survey {self.name}.")
            return False
        del data[dtype]
        self.config.update({"data": data})
        write_dataset_config(self.name, self.config)
        return True

    @check_overload
    def get_from(
        self, dtypes: list, region_overlaps: list, *args, **kwargs
    ) -> dict[str, Any]:
        return_types = []
        new_data = {}
        if not isinstance(region_overlaps[0], str):
            regnames = set([r.name for r in region_overlaps])
        else:
            regnames = set(region_overlaps)

        self.load_handlers()
        data = self.config.get("data", {})
        for dtype in dtypes:
            try:
                _ = data[dtype]
                return_types.append(dtype)
            except KeyError:
                raise MissingDataError(
                    f"Data of type {dtype} not found for dataset {self.name}!"
                )

        cache = get_cache(self.name)
        cached_data = cache.get(regnames, dtypes)

        for dtype in return_types:
            if dtype in cached_data:
                regions_to_get = [r for r in regnames if r not in cached_data[dtype]]
            else:
                regions_to_get = regnames

            if len(regions_to_get) != 0:
                data_ = self._handlers[dtype].get_data(regions_to_get, *args, **kwargs)
                new_data.update({dtype: data_})

        if len(new_data) != 0:
            cache.add(new_data)
        storage = {}
        for dtype in return_types:
            cached_data_of_dtype = cached_data.get(dtype, {})
            new_data_of_dtype = new_data.get(dtype, {})
            found_regions = set(cached_data_of_dtype.keys()) | set(
                new_data_of_dtype.keys()
            )
            missing_regions = regnames - found_regions
            if len(missing_regions) != 0:
                raise MissingDataError(
                    f"Could not find data for regions {missing_regions} of"
                    f" type {dtype}"
                )

            storage.update({dtype: {**cached_data_of_dtype, **new_data_of_dtype}})

        return storage

    @check_overload
    def get_data(
        self,
        dtypes: list,
        query_region: BaseRegion,
        region_overlaps: list,
        *args,
        **kwargs,
    ) -> dict[str, Any]:
        """
        Get data of a specificed type
        The manager is responsible for finding the path, and the giving
        it to the handlers
        """

        storage = self.get_from(dtypes, region_overlaps, *args, **kwargs)
        storage = self.parse_data(storage, *args, **kwargs)
        return storage

    def parse_data(self, data, *args, **kwargs) -> dict[str, Any]:
        return_data = {}
        for dtype, values in data.items():
            # Now, we process into useful objects and filter further
            if data is None:
                logger.error(f"Unable to find data of type {dtype}")
                continue
            try:
                objs = list(values.values())
                data_obj = objs[0].combine(objs)
                return_data.update({dtype: data_obj})
            except IndexError:
                return_data.update({dtype: None})
        return return_data



from heinlein import api

from .config import get_option, set_option
from .dataset import load_dataset
from .region import Region

__version__ = "0.9.1"
__all__ = ["load_dataset", "get_option", "set_option", "Region"]
__all__.extend(api.__all__)



from functools import singledispatchmethod

import astropy.units as u
import healpy

from .region import BaseRegion


def partition_regions(
    regions: list[BaseRegion], nside: int
) -> dict[str, list[BaseRegion]]:
    output = {}
    for region in regions:
        pixels = query_healpix(region, nside)
        for pixel in pixels:
            pixel_output = output.get(pixel, [])
            pixel_output.append(region)
            output[pixel] = pixel_output
    return output


def query_healpix(region: BaseRegion, nside: int):
    lon, lat = region.bounding_box.to_lonlat()
    lon, lat = lon[:-1], lat[:-1]
    vecs = healpy.ang2vec(lon, lat, lonlat=True)
    return healpy.query_polygon(nside, vecs, inclusive=True)


class Footprint:
    """
    The footprint class is a container for a set of regions. It is used to represent
    the geometry of the survey, which is used to accelerate queries.

    The initial querying for regions is done using Healpix. When a footprint is created,
    it determines which healpix pixels overlap with each of its regions. The first step
    in querying is then to determine which healpix pixels overlap with the region being
    queried, and return
    the regions that overlap with those pixels.
    """

    def __init__(self, regions: list[BaseRegion], nside=64, *args, **kwargs):
        self._regions = partition_regions(regions, nside)
        self._nside = nside

    @singledispatchmethod
    def get_overlapping_regions(self, query_region: BaseRegion) -> list[BaseRegion]:
        """
        Returns a list of regions that overlap with the given region
        """
        pixels = query_healpix(query_region, self._nside)
        output = []
        for pixel in pixels:
            overlap = self._regions.get(pixel, [])
            overlap = filter(lambda region: region.intersects(query_region), overlap)
            output.extend(overlap)
        return output

    @get_overlapping_regions.register
    def _(self, region: list) -> list[list[BaseRegion]]:
        return [self.get_overlapping_regions(r) for r in region]

    def get_overlapping_region_names(
        self, region: BaseRegion | list[BaseRegion]
    ) -> list[str]:
        """
        Returns a list of region names that overlap with the given region
        """
        overlaps = self.get_overlapping_regions(region)
        if isinstance(region, list):
            return [[r.name for r in o] for o in overlaps]
        return [r.name for r in overlaps]

    def sample(self, n: int = 1, tolerance: u.Quantity = None, *args, **kwargs) -> list:
        """
        Return n random points from the footprint
        """
        if n == 1:
            return self._sampler.sample(tolerance=tolerance)
        else:
            return [self._sampler.sample(tolerance=tolerance) for _ in range(n)]

    def partition_by_region(
        self, samples: list[BaseRegion]
    ) -> dict[str, list[BaseRegion]]:
        """
        Given a list of regions, partition based on the regions they overlap
        with in this footprint. Returns a dictionary by region name. Samples
        that overlap with multiple regions are placed in a key that is a
        concatenation of the regions they overlap with.
        """
        overlaps = self.get_overlapping_region_names(samples)
        partitions = {}

        for i, sample in enumerate(samples):
            overlap = overlaps[i]
            if len(overlap) == 1:
                okey = overlap[0]
            else:
                overlap.sort()
                okey = "/".join(overlap)
            try:
                partitions[okey].append(sample)
            except KeyError:
                partitions[okey] = [sample]
        return partitions



from __future__ import annotations

import logging
from functools import partial
from typing import Any, Optional, Union

import astropy.units as u
from astropy.coordinates import SkyCoord

from heinlein.dataset.extension import get_extension, load_extensions
from heinlein.manager import get_manager
from heinlein.manager.cache import clear_cache
from heinlein.manager.manager import DataManager
from heinlein.region import BaseRegion, Region
from heinlein.region.footprint import Footprint

logger = logging.getLogger("Dataset")


def load_dataset(name: str) -> Dataset:
    """
    Load a dataset by name. This method shold always be used. Do
    not instantiate the dataset class directly.
    """
    manager = get_manager(name)
    ds = Dataset(manager)
    ds = setup_dataset(ds)
    return ds


def load_current_config(name: str):
    manager = get_manager(name)
    return manager.config


def setup_dataset(dataset: Dataset):
    """
    Searches for an external implementation of the datset
    This is used for datasets with specific needs (i.e. specific surveys)
    """
    external = dataset.manager.external

    if external and dataset.manager.external is None:
        raise NotImplementedError(
            f"No implementation code found for datset {dataset.name}"
        )

    try:
        get_regions = dataset.manager.get_external("load_regions")
    except KeyError:
        raise NotImplementedError(
            f"Dataset {dataset.name} does not have a setup method!"
        )

    regions = get_regions()
    dataset.footprint = Footprint(regions)
    load_extensions(dataset)
    return dataset


class Dataset:
    """
    The Dataset class is the core of heinlein's user interface. It contains routines
    for querying data from the underlying dataset.

    The dataset class should not be created directly. Instead use "load_dataset."

    """

    def __init__(self, manager: DataManager, *args, **kwargs):
        self.manager = manager
        self._extensions = {}
        self._parameters = {}

    @property
    def name(self):
        return self.manager.name

    def __getattr__(self, key__):
        """
        I went through a phase where I thought defining magic methods all over the place
        was really cool. I've since learned that it sometimes creates more complexity
        than it is worth, but this is one case where it makes sense.

        There are certain methods that need to be defined or overloaded for specific
        datasets. This method allows that to be largely transparent to the end user,
        without cluttering the main class with a bunch of extra methods.
        """

        try:
            f = get_extension(self.name, key__)
            return partial(f, self)
        except KeyError:
            raise AttributeError(
                f"{type(self).__name__} object has no attribute '{key__}'"
            )

    def cone_search(
        self, center: tuple | SkyCoord, radius: u.Quantity, *args, **kwargs
    ) -> dict[str, Any]:
        """
        Perform a cone search on the dataset. Will return a dictionary of format:
        {"datta_type": data}

        The center can be passed in as a sky coordinate or a tuple of (ra, dec). If
        a tuple, it is assumed to be in degrees.
        """
        reg = Region.circle(center=center, radius=radius)
        return self.get_data_from_region(reg, *args, **kwargs)

    def box_search(
        self,
        center: tuple | SkyCoord,
        width: u.Quantity,
        height: Optional[u.Quantity] = None,
        *args,
        **kwargs,
    ) -> dict[str, Any]:
        """
        Perform a box search on the dataset. Will return a dictionary of format:
        {"datta_type": data}
        """
        if isinstance(center, tuple):
            center = SkyCoord(*center, unit="deg")
        if height is None:
            height = width
        minx = center.ra - width / 2
        maxx = center.ra + width / 2
        miny = center.dec - height / 2
        maxy = center.dec + height / 2
        reg = Region.box([minx, miny, maxx, maxy])
        return self.get_data_from_region(reg, *args, **kwargs)

    def get_data_from_region(
        self,
        query_region: BaseRegion,
        dtypes: Union[str, list] = "catalog",
        *args,
        **kwargs,
    ) -> dict[str, Any]:
        """
        Get data of of the given types from a provided region. In general,
        it is better to use cone_search or box_search so you don't have to
        worry about constructing the region yourself.

        Paramaters:

        region <BaseRegion> heinlein Region object
        dtypes <str> or <list>: list of data types to return

        """
        get_overlaps = self.manager.get_external("get_overlapping_regions")
        if get_overlaps is not None:
            overlaps = get_overlaps(self, query_region)
        else:
            overlaps = self.footprint.get_overlapping_regions(query_region)
        if len(overlaps) == 0:
            print("Error: No objects found in this region!")
            return

        data = {}
        if isinstance(dtypes, str):
            dtypes = [dtypes]

        data = self.manager.get_data(dtypes, query_region, overlaps, *args, **kwargs)
        return_data = {}
        for dtype, obj_ in data.items():
            try:
                return_data.update({dtype: obj_.get_data_from_region(query_region)})
            except AttributeError:
                return_data.update({dtype: obj_})

        return return_data

    def get_data_from_samples(
        self,
        samples,
        sample_type="cone",
        sample_dimensions=45 * u.arcsec,
        dtypes=["catalog"],
        *args,
        **kwargs,
    ):
        """
        Often, we want to get many many samples in a row.
        is that the cache will very quickly blow up if the samples cover may
        of the survey's subregions. This methord returns a generator that
        yields samples from the survey. It orders them such that it can
        only load a few subregions at a time, then dumps them from the cache when
        they are done.
        """
        if sample_type != "cone":
            raise NotImplementedError("Only cone sampling is currently supported")

        samples = [Region.circle(center=s, radius=sample_dimensions) for s in samples]

        partitions = self.footprint.partition_by_region(samples)
        counts = {p: p.count("/") + 1 for p in partitions.keys()}
        done = set()

        for reg_key in partitions.keys():
            if counts[reg_key] == 1:
                continue
            regs_to_get = reg_key.split("/")
            s = partitions[reg_key]
            for s_ in s:
                yield (s_, self.get_data_from_region(s_, dtypes))
            for reg_ in regs_to_get:
                # Now, get the samples that fall into a single one of the regions
                if reg_ in done or reg_ not in partitions:
                    continue
                done.add(reg_)
                for s_ in partitions[reg_]:
                    yield (s_, self.get_data_from_region(s_, dtypes))
            # Now, we dump the data from those regions
            done.add(reg_key)
            clear_cache(self.name)
        # Now we go through all the samples that fall in a single survey region
        # that were NOT covered before
        remaining = set(partitions.keys()) - done
        for reg_key in remaining:
            s = partitions[reg_key]
            for s_ in s:
                yield (s_, self.get_data_from_region(s_, dtypes))
            clear_cache(self.name)

    def clear_cache(self):
        """
        Clear the cache for this dataset, useful if you are sampling a lot and need
        to clear up memory.
        """
        clear_cache(self.name)
        return



from pathlib import Path
from types import ModuleType

from . import catalog, mask


def get_file_handlers(dtypes: list, config: dict, external: dict, *args, **kwargs):
    if external is not None:
        external_handlers = get_external_handlers(config, external)
    else:
        external_handlers = {dtype: None for dtype in dtypes}
    handlers_ = {}
    all_dconfig = config.get("dconfig", {})
    data = config.get("data", {})
    for dtype in dtypes:
        dconfig = all_dconfig.get(dtype, {})
        if external_handlers[dtype] is not None:
            cl = external_handlers[dtype](Path(data[dtype]), dconfig)
            handlers_.update({dtype: cl})
        elif dtype == "catalog":
            cl = catalog.get_catalog_handler(config, dconfig)
            handlers_.update({dtype: cl})
        elif dtype == "mask":
            cl = mask.get_mask_handler(config, dconfig)
            handlers_.update({dtype: cl})
    return handlers_


def get_external_handlers(config: dict, external: ModuleType):
    output = {}
    known_dtypes = list(config.get("data", {}).keys())

    for dtype in known_dtypes:
        function_key = f"{dtype.capitalize()}Handler"
        output.update({dtype: external.get(function_key, None)})
    return output



import click

from . import cmds


@click.group()
def cli():
    pass


cli.add_command(cmds.add)
cli.add_command(cmds.remove)
cli.add_command(cmds.get)
cli.add_command(cmds.prep)

if __name__ == "__main__":
    cli()



import os
import warnings
from typing import Any

from pydantic import ValidationError

from heinlein.errors import HeinleinConfigError

from .config import HeinleinConfig

try:
    _HEINLEIN_CONFIG = HeinleinConfig()
except ValidationError as e:
    bad_values = {e["loc"][0]: e["msg"] for e in e.errors()}
    msg = (
        "heinlein recieved invalid configuration values from the "
        "environment and will use defaults instead. Invalid values:\n"
    )
    msg += "\n".join(f"{k}: {v}" for k, v in bad_values.items())
    warnings.warn(msg)
    for k in bad_values.keys():
        os.environ.pop(f"HEINLEIN_{k}", None)
    _HEINLEIN_CONFIG = HeinleinConfig()


def get_option(option: str) -> Any:
    canonical_name = option.upper()
    try:
        return getattr(_HEINLEIN_CONFIG, canonical_name)
    except AttributeError:
        raise HeinleinConfigError.doesnt_exist(option)


def set_option(option: str, value: Any) -> bool:
    canonical_name = option.upper()
    if not hasattr(_HEINLEIN_CONFIG, canonical_name):
        raise HeinleinConfigError.doesnt_exist(option)
    setattr(_HEINLEIN_CONFIG, canonical_name, value)
    return True



from inspect import getmembers


class dataset_extension:
    def __init__(self, function):
        self.function = function
        self.extension = True
        self.name = function.__name__

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)


KNOWN_EXTENSIONS = {}


def load_extensions(dataset, *args, **kwargs):
    """
    Loads extensions for the particular dataset. These are defined externally
    """
    if dataset.name in KNOWN_EXTENSIONS:
        return

    ext_objs = list(
        filter(
            lambda f: isinstance(f[1], dataset_extension),
            getmembers(dataset.manager.external),
        )
    )

    extensions = {f[0]: f[1] for f in ext_objs}
    KNOWN_EXTENSIONS[dataset.name] = extensions


def get_extension(dataset, name):
    """
    Get the extension object for the dataset
    """
    return KNOWN_EXTENSIONS[dataset][name]



import warnings
from abc import ABC, abstractmethod
from functools import singledispatchmethod
from gettext import Catalog
from typing import Iterable

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.io.fits import HDUList
from astropy.nddata import Cutout2D
from astropy.nddata.utils import NoOverlapError
from astropy.utils.exceptions import AstropyWarning
from astropy.wcs import WCS, utils
from shapely import get_num_geometries
from shapely.geometry import MultiPoint
from shapely.geometry.base import BaseGeometry
from shapely.strtree import STRtree
from spherical_geometry.vector import lonlat_to_vector

from heinlein.dtypes.dobj import HeinleinDataObject
from heinlein.region import BaseRegion

warnings.simplefilter("ignore", category=AstropyWarning)


def get_mask_objects(input_list, *args, **kwargs):
    output_data = np.empty(len(input_list), dtype="object")
    if all(isinstance(obj, BaseGeometry) for obj in input_list):
        return [_regionMask(input_list)]
    for index, obj in enumerate(input_list):
        if type(obj) == HDUList:
            # if kwargs.get("pixarray", False):
            #    output_data[index] = _pixelArrayMask(obj, *args, **kwargs)
            # else:
            output_data[index] = _fitsMask.from_hdu(obj, *args, **kwargs)
        elif type(obj) == np.ndarray:
            if isinstance(obj[0], BaseRegion):
                output_data[index] = _regionMask(obj)

    return output_data


class Mask(HeinleinDataObject):
    def __init__(self, masks=[], *args, **kwargs):
        """
        Masks are much less regular than catalogs. There are many different
        formats, and some surveys mix formats. The "Mask" object is a wraper class
        that handles interfacing with all these different formats.
        """
        self._masks = get_mask_objects(masks, *args, **kwargs)
        self._check_filter = False
        self._can_filter = False

    @classmethod
    def combine(cls, masks, *args, **kwargs):
        all_masks = [m._masks for m in masks]
        all_masks = np.hstack(all_masks)
        return cls.from_masks(all_masks)

    def __len__(self):
        return len(self._masks)

    @classmethod
    def from_masks(cls, masks, *args, **kwargs):
        m = cls()
        m._masks = masks
        return m

    def estimate_size(self) -> int:
        return sum([mask.estimate_size() for mask in self._masks])

    def mask(self, catalog: Catalog, *args, **kwargs):
        for mask in self._masks:
            catalog = mask.mask(catalog)
            if len(catalog) == 0:
                return catalog
        return catalog

    def append(self, other):
        if len(other) == 0:
            return self

        masks = np.empty(1 + len(other), dtype=object)
        for index, m_ in enumerate(other):
            masks[index + 1] = m_._masks
        masks[0] = self._masks
        all_masks = np.hstack(masks)
        return Mask.from_masks(all_masks)

    def get_data_from_region(self, region: BaseRegion):
        if not self._check_filter:
            self._can_filter = any(
                [hasattr(mask, "get_data_from_region") for mask in self._masks]
            )
        if not self._can_filter:
            raise AttributeError

        return_vals = []
        for mask in self._masks:
            try:
                submask = mask.get_data_from_region(region)
                if submask is not None:
                    return_vals.append(submask)
            except AttributeError:
                return_vals.append(mask)
        return Mask.from_masks(return_vals)


class _mask(ABC):
    def __init__(self, mask, *args, **kwargs):
        """
        Basic mask object. Handles a single mask of an arbitary type.
        """
        self._mask = mask

    @abstractmethod
    def mask(self, *args, **kwargs):
        """
        For subclasses, the actual masking logic will be implmeneted here.
        """
        pass

    @abstractmethod
    def estimate_size(self) -> int:
        """
        Estimate the size of the mask in bytes.
        """
        pass


class _mangleMask(_mask):
    def __init__(self, mask, *args, **kwargs):
        """
        Implementation for masks in mangle format
        """
        super().__init__(mask)

    @singledispatchmethod
    def mask(self, catalog: Catalog, *args, **kwargs):
        coords = catalog["coordinates"]
        contains = self._check(coords)
        return catalog[~contains]

    @mask.register
    def _(self, coords: SkyCoord):
        contains = self._check(coords)
        return coords[~contains]

    def _check(self, coords):
        ra = coords.ra.to("deg").value
        dec = coords.dec.to("deg").value
        contains = self._mask.contains(ra, dec)
        return contains

    def estimate_size(self) -> int:
        return 0


class _pixelArrayMask(_mask):
    def __init__(self, mask, mask_key, *args, **kwargs):
        super().__init__(mask)
        self._wcs = WCS(mask[0].header)
        self._mask_key = mask_key
        self._init_pixel_array()

    def _init_pixel_array(self, *args, **kwargs):
        mask = self._mask[self._mask_key].data

        pixels = np.where(mask > 0)
        pixel_coords = np.ones(mask.shape, dtype=bool)
        pixel_coords[pixels] = False
        self._mask.close()
        self._mask = pixel_coords

    def _check(self, coords, *args, **kwargs):
        pix_coords = self._wcs.world_to_pixel(coords)
        shape = self._mask.shape
        x = np.round(pix_coords[0], 0).astype(int)
        y = np.round(pix_coords[1], 0).astype(int)
        x_lims = (x < 0) | (x >= shape[0])
        y_lims = (y < 0) | (y >= shape[1])
        m_ = x_lims | y_lims  # These objects are outside this particular mask

        # plt.scatter(coords.ra, coords.dec, transform=ax.get_transform('world'))

        unmasked_objects = np.zeros(len(coords), dtype=bool)
        unmasked_objects[~m_] = self._mask[x[~m_], y[~m_]]

        return unmasked_objects

    def estimate_size(self) -> int:
        return self._mask.nbytes

    @singledispatchmethod
    def mask(self, catalog: Catalog, *args, **kwargs):
        coords = catalog["coordinates"]
        mask = self._check(coords)
        return catalog[mask]

    @mask.register
    def _(self, coords: SkyCoord):
        mask = self._check(coords)
        return coords[mask]


class _fitsMask(_mask):
    def __init__(self, mask, wcs, mask_data):
        super().__init__(mask)
        self._wcs = wcs
        self._mask = mask_data

    @classmethod
    def from_hdu(cls, mask, mask_key, pixarray=False, *args, **kwargs):
        """
        If the mask is stored in a fits file, we assume we can
        get the WCS info from the header in HDU 0, but we need to
        know where the actual mask is located, so we pass a key.
        A fits mask assumes masked pixels have a value > 0
        """

        wcs = WCS(mask[0].header)
        mask_key = mask_key
        mask_plane = mask[mask_key].data
        return cls(mask, wcs, mask_plane)

    @classmethod
    def from_cutout(cls, cutout):
        wcs = cutout.wcs
        data = cutout.data
        return cls(cutout, wcs, data)

    def estimate_size(self) -> int:
        return self._mask.nbytes

    def _check(self, coords):
        y, x = utils.skycoord_to_pixel(coords, self._wcs)
        # The order of numpy axes is the opposite of the order
        # in fits images. All the data is being stored in
        # arrays, so we have to flip things here.
        x = np.round(x, 0).astype(int)
        y = np.round(y, 0).astype(int)
        masked = np.ones(len(x), dtype=bool)

        x_limit = self._mask.shape[0]
        y_limit = self._mask.shape[1]

        negative_check = (x < 0) | (y < 0)
        # Note: We already inverted indices above
        x_limit_check = x >= x_limit
        y_limit_check = y >= y_limit
        to_skip = negative_check | x_limit_check | y_limit_check

        masked[to_skip] = False
        pixel_values = self._mask[x[~to_skip], y[~to_skip]]
        masked[~to_skip] = pixel_values > 0
        return ~masked

    @singledispatchmethod
    def mask(self, catalog: Catalog, *args, **kwargs):
        coords = catalog["coordinates"]
        mask = self._check(coords)
        return catalog[mask]

    def get_data_from_region(self, region: BaseRegion):
        bounds = [b * u.deg for b in region.bounds]
        center = SkyCoord((bounds[0] + bounds[2]) / 2, (bounds[1] + bounds[3]) / 2)
        size = (bounds[2] - bounds[0], bounds[3] - bounds[1])

        try:
            cutout = Cutout2D(self._mask, center, size, wcs=self._wcs, copy=True)
        except NoOverlapError:
            return None

        return _fitsMask.from_cutout(cutout)

    @mask.register
    def _(self, coords: SkyCoord):
        mask = self._check(coords)
        return coords[mask]


class _regionMask(_mask):
    def __init__(self, mask: Iterable[BaseRegion], *args, **kwargs):
        """
        Implementation for masks defined as regions.
        """
        super().__init__(mask)
        self._init_tree()

    def _init_tree(self):
        self._geo_list = np.fromiter(self._mask, dtype=object)
        self._geo_tree = STRtree(self._geo_list)

    @singledispatchmethod
    def mask(self, catalog: Catalog):
        cmask = self.generate_mask(catalog["coordinates"])
        return catalog[cmask]

    @mask.register
    def _(self, coords: SkyCoord):
        cmask = self.generate_mask(coords)
        return coords[cmask]

    def estimate_size(self) -> int:
        return self._geo_list.nbytes

    def generate_mask(self, coords: SkyCoord):
        ra = coords.ra.to_value("deg")
        dec = coords.dec.to_value("deg")
        points = MultiPoint(np.dstack(lonlat_to_vector(ra, dec))[0])
        mask = self._check(points)
        return mask

    def _check(self, points):
        mask = np.ones(get_num_geometries(points), dtype=bool)
        for index, p in enumerate(points.geoms):
            a = self._geo_tree.query(p)
            for geo in a:
                if self._geo_list[geo].contains(p):
                    mask[index] = False
                    break
        return mask



from pathlib import Path

import pandas as pd
from sqlalchemy import MetaData, create_engine, text

from heinlein.dtypes.catalog import load_config
from heinlein.manager.manager import get_manager, initialize_dataset


def database_from_csvs(dataset_name: str, csvs: list[Path]):
    try:
        manager = get_manager(dataset_name)
    except FileNotFoundError:
        initialize_dataset(dataset_name)
        manager = get_manager(dataset_name)
    try:
        catalog_config = manager.config["dconfig"]["catalog"]
    except KeyError:
        raise KeyError("Catalog configuration not found in dataset configuration")

    db_path = csvs[0].parent / f"{dataset_name}.db"
    engine = create_engine(f"sqlite:///{db_path}").connect()

    region_key = catalog_config["region"]
    index_key = catalog_config["index_key"]

    meta = MetaData()
    column_config = load_config()["columns"]
    for csv in csvs:
        data = pd.read_csv(csv)
        for name, aliases in column_config.items():
            for alias in aliases:
                if alias in data.columns:
                    data = data.rename({alias: name})

        if region_key not in data.columns:
            raise KeyError(f"Region key {region_key} not found in {csv.name}")
        if index_key not in data.columns:
            raise KeyError(f"Index key {index_key} not found in {csv.name}")
        partitions = data.groupby(region_key)
        for rname, partition in partitions:
            meta.reflect(engine)
            tables_in_db = meta.tables.keys()
            if rname not in tables_in_db:
                partition.to_sql(
                    rname,
                    engine,
                    if_exists="fail",
                )

            else:
                partition.to_sql(
                    f"{rname}_temp",
                    engine,
                    if_exists="fail",
                )

                query = text(
                    f"""
                    INSERT INTO '{rname}'
                    SELECT * FROM '{rname}_temp'
                    WHERE {index_key} NOT IN (SELECT {index_key} FROM '{rname}')
                    """
                )
                engine.execute(query)
                engine.execute(text(f"DROP TABLE '{rname}_temp'"))
                engine.commit()
    return db_path


def regularize_sqlite_databse(database_path: Path):
    engine = create_engine(f"sqlite:///{database_path}").connect()
    column_config = load_config()["columns"]
    meta = MetaData()
    meta.reflect(engine)
    tables_in_db = meta.tables.keys()

    for table in tables_in_db:
        table = meta.tables[table]
        for name, aliases in column_config.items():
            for alias in aliases:
                if alias in table.columns:
                    engine.execute(
                        f"""
                        ALTER TABLE {table} RENAME COLUMN {alias} TO {name}
                        """
                    )



from pydantic import ByteSize, Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class HeinleinConfig(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="HEINLEIN_", validate_assignment=True)
    CACHE_ENABLED: bool = Field(False)
    CACHE_SIZE: ByteSize = Field(4e9, ge=1e9)



from pathlib import Path

import appdirs

base = Path(__file__).parents[0]

data_dir = Path(appdirs.user_data_dir("heinlein"))

INSTALL_DIR = base
MAIN_CONFIG_DIR = base / "config"

BASE_DATASET_CONFIG_DIR = MAIN_CONFIG_DIR / "datasets"
BASE_DATASET_CONFIG = BASE_DATASET_CONFIG_DIR / "surveys.json"

BUILTIN_DTYPES = MAIN_CONFIG_DIR / "dtypes" / "dtypes.json"

DATASET_CONFIG_DIR = data_dir / "datasets"
DATASET_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
MAIN_DATASET_CONFIG = DATASET_CONFIG_DIR / "surveys.json"
__all__ = [
    "BASE_DATASET_CONFIG_DIR",
    "MAIN_DATASET_CONFIG",
    "DATASET_CONFIG_DIR",
    "MAIN_DATASET_CONFIG",
]



import astropy.units as u
import numpy as np
from shapely import GeometryCollection, Point


class Sampler:
    def __init__(self, footprint: GeometryCollection, *args, **kwargs):
        """
        Base class for samplers
        """
        self.setup(footprint)

    def setup(self, footprint: GeometryCollection, *args, **kwargs) -> None:
        self._footprint = footprint
        bounds = self._footprint.bounds
        ra1, ra2 = bounds[0], bounds[2]
        dec1, dec2 = bounds[1], bounds[3]
        ra_range = (min(ra1, ra2), max(ra1, ra2))
        dec_range = (min(dec1, dec2), max(dec1, dec2))
        phi_range = np.radians(ra_range)
        # Keeping everything in radians for simplicity
        # Convert from declination to standard spherical coordinates
        theta_range = (90.0 - dec_range[0], 90.0 - dec_range[1])
        # Area element on a sphere is dA = d(theta)d(cos[theta])
        # Sampling uniformly on the surface of a sphere means sampling uniformly
        # Over cos theta
        costheta_range = np.cos(np.radians(theta_range))
        self._low_sampler_range = [phi_range[0], costheta_range[0]]
        self._high_sampler_range = [phi_range[1], costheta_range[1]]
        self._sampler = np.random.default_rng()

    def sample(self, n=1, tolerance: u.Quantity = None, *args, **kwargs) -> tuple:
        vals = self._sampler.uniform(self._low_sampler_range, self._high_sampler_range)
        ra = np.degrees(vals[0])
        theta = np.degrees(np.arccos(vals[1]))
        dec = 90 - theta

        shapely_point = Point(ra, dec)

        if self._footprint.contains(shapely_point):
            if tolerance is None:
                return ra, dec
            region = shapely_point.buffer(tolerance.to(u.deg).value)
            if self._footprint.contains_properly(region):
                return ra, dec
        return self.sample(n)



import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from spherical_geometry.polygon import SingleSphericalPolygon

from heinlein.region import sampling
from heinlein.region.base import BaseRegion, create_bounding_box
from heinlein.utilities.utilities import initialize_grid


def parse_angle_list(angles: list[u.Quantity]):
    """
    Parse a list of angles into floats. In units of degrees.
    """
    angles = list(
        map(
            lambda angle: angle.to(u.deg).value
            if isinstance(angle, u.Quantity)
            else angle,
            angles,
        )
    )
    # Regularize RAs to be between 0 and 360
    angles[0] = angles[0] % 360
    angles[2] = angles[2] % 360
    # Regularize DECs to be between -90 and 90
    # -92 should map to 88
    angles[1] = (angles[1] + 90) % 180 - 90
    angles[3] = (angles[3] + 90) % 180 - 90
    return angles


def create_reflections(bounds: tuple) -> list[tuple]:
    """
    Bounds checking is complicated when region straddles the 0/360 line
    or the poles. This function creates the reflections of the region
    across the 0/360 line and the poles if necessary.

    Because of how the angles are parsed, we can always guarantee that
    RAs are between 0 and 360 and DECs are between -90 and 90
    """
    output_bounds = [bounds]
    ra_min, dec_min, ra_max, dec_max = bounds
    if ra_min > ra_max:
        output_bounds.append((ra_min - 360, dec_min, ra_max, dec_max))
        output_bounds.append((ra_min, dec_min, ra_max + 360, dec_max))
    if dec_min > dec_max:
        new_bounds = []
        for bound_set in output_bounds:
            ra_min, dec_min, ra_max, dec_max = bound_set
            new_bounds.append((ra_min, dec_min + 180, ra_max, dec_max))
            new_bounds.append((ra_min, dec_min, ra_max, dec_max - 180))
        output_bounds.extend(new_bounds)
    return output_bounds


class Region:
    @staticmethod
    def circle(
        center: SkyCoord | tuple,
        radius: u.Quantity | float,
        name: str = None,
        *args,
        **kwargs,
    ):
        """
        Return a circular region. Centered on center, radius of `radius`
        The "center" is anything that can be parsed to a SkyCoord_
        If no units provided, will default to degrees
        """
        if not isinstance(center, SkyCoord):
            center = SkyCoord(*center, unit="deg")
        if not isinstance(radius, u.Quantity):
            radius = radius * u.deg

        return CircularRegion(center, radius, name, *args, **kwargs)

    @staticmethod
    def box(bounds: tuple, name: str = None):
        try:
            if len(bounds) != 4:
                raise ValueError("Invalid bounds: must be 4 values")
        except TypeError:
            raise ValueError("Invalid bounds: must be a tuple of 4 values")
        bounds_degree = parse_angle_list(bounds)
        return BoxRegion(bounds_degree, name)


class PolygonRegion(BaseRegion):
    def __init__(
        self, points: list[tuple], inside=None, name: str = None, *args, **kwargs
    ):
        """
        Polygon region. Accepts a list of points for initialization.

        parameters:

        points <list>: A list of points that define the polygon
        name <str>: a name for the region (optional)
        """
        ra = [point[0] for point in points]
        dec = [point[1] for point in points]
        polygon = SingleSphericalPolygon.from_lonlat(ra, dec, inside)
        ra_min = min(ra)
        ra_max = max(ra)
        dec_min = min(dec)
        dec_max = max(dec)
        bounds = (ra_min, dec_min, ra_max, dec_max)
        super().__init__(polygon, bounds, name, "PolygonRegion", *args, **kwargs)
        self._sampler = None

    def contains(self, point: SkyCoord) -> bool:
        return self.spherical_geometry.contains_radec(
            point.ra.deg, point.dec.deg, degrees=True
        )


class BoxRegion(BaseRegion):
    def __init__(self, bounds: tuple, name: str = None, *args, **kwargs):
        """
        Basic general-shape region object.

        Parameters:

        polygon <spherical_geometry.SingleSphericalPolygon>: The spherical polygon
        representing the region

        name <str>: a name for the region (optional)
        """
        polygon = create_bounding_box(bounds)
        super().__init__(polygon, bounds, name, "BoxRegion", *args, **kwargs)
        self._sampler = None

    def contains(self, point: SkyCoord) -> bool:
        """
        Check if a point is contained within the region

        Spherical gemoetry's contain method is much too slow
        for large numbers of points. Eventually I would like
        to migrate to Googles S2 library for this, but for now
        we do basic bounds checking.
        """
        bounds_to_check = create_reflections(self.bounds)
        mask = np.zeros((len(point), 4 * len(bounds_to_check)), dtype=bool)
        bounds = np.array(bounds_to_check).flatten()
        ra_mins = bounds[::4]
        dec_mins = bounds[1::4]
        ra_maxs = bounds[2::4]
        dec_maxs = bounds[3::4]
        ra_arr = point.ra.deg[:, np.newaxis]  #
        dec_arr = point.dec.deg[:, np.newaxis]
        mask[:, 0::4] = ra_arr >= ra_mins
        mask[:, 1::4] = dec_arr >= dec_mins
        mask[:, 2::4] = ra_arr <= ra_maxs
        mask[:, 3::4] = dec_arr <= dec_maxs
        # At this point we have a 2d array of shape (len(point), 4*len(bounds_to_check))
        # Each row is a point, each group of four entires in the row is a comparison to
        # one of the sets of bounds
        # If any group of four in the row is all true, the point is in the region
        mask = mask.reshape(len(point), len(bounds_to_check), 4)
        mask = mask.all(axis=2).any(axis=1)
        return mask

    def generate_circular_tile(self, radius, *args, **kwargs):
        """
        Return a circular tile, drawn randomly from the region.
        """
        if getattr(self, "_sampler", None) is None:
            self._get_sampler()
        return self._sampler.get_circular_sample(radius)

    def generate_circular_tiles(self, radius, n, *args, **kwargs):
        if self._sampler is None:
            self._get_sampler()
        return self._sampler.get_circular_samples(radius, n)

    def _get_sampler(self, *args, **kwargs):
        self._sampler = sampling.Sampler(self)

    def initialize_grid(self, density=1000, *args, **kwargs):
        bounds = self.sky_geometry.bounds
        area = self.sky_geometry.area
        coords = initialize_grid(bounds, area, density)
        return coords

    generate_grid = initialize_grid


class CircularRegion(BaseRegion):
    def __init__(
        self, center: SkyCoord, radius: u.Quantity, name: str = None, *args, **kwargs
    ):
        """
        Circular region. Accepts point-radius for initialization.

        parameters:

        center <SkyCoord>: The center of the region
        radius <astropy.units.quantity>: The radius of the region
        name <str>: a name for the region (optional)
        """
        self._center = (center.ra.deg, center.dec.deg)
        self._radius = radius.to_value("deg")
        self._skypoint = center
        self._unitful_radius = radius
        geometry = SingleSphericalPolygon.from_cone(
            *self._center,
            self._radius,
        )
        bounds = (
            self._center[0] - self._radius,
            self._center[1] - self._radius,
            self._center[0] + self._radius,
            self._center[1] + self._radius,
        )
        super().__init__(geometry, bounds, name, "CircularRegion", *args, **kwargs)

    @property
    def center(self) -> SkyCoord:
        return self._skypoint

    @property
    def radius(self) -> u.quantity:
        return self._unitful_radius

    def contains(self, points: SkyCoord) -> bool | np.ndarray:
        distance = self.center.separation(points)
        return distance <= self.radius
